@@ -70,11 +70,13 @@ def squeeze(sequence):
     return result
 
 
-def get_line_confidence(line, labels):
-    log_probs = line.get_full_logprobs()
+def get_line_confidence(line, labels, aligned_letters=None, log_probs=None):
+    if log_probs is None:
+        log_probs = line.get_full_logprobs()
 
-    alignment = align_text(-log_probs, labels, log_probs.shape[1] - 1)
-    alignment = np.concatenate([alignment, [1000]])
+    if aligned_letters is None:
+        aligned_letters = align_text(-log_probs, labels, log_probs.shape[1] - 1)
+    alignment = np.concatenate([aligned_letters, [1000]])
 
     probs = np.exp(log_probs)
     last_border = 0
@@ -85,9 +87,13 @@ def get_line_confidence(line, labels):
         pos_probs = probs[last_border: next_border]
         masked_probs = np.copy(pos_probs)
         masked_probs[:, label] = 0
+        if i > 0:
+            masked_probs[:, labels[i-1]] = 0
+        if i + 1 < len(labels):
+            masked_probs[:, labels[i+1]] = 0
         other_prob = masked_probs[:, :-1].max()
-        confidences[i] = label_prob - other_prob
+        confidences[i] = max(0, label_prob - other_prob)
         last_border = next_border
 
-    confidences = confidences / 2 + 0.5
+    #confidences = confidences / 2 + 0.5
     return confidences
