@@ -1,13 +1,15 @@
 import numpy as np
 
+import logging
 from multiprocessing import Pool
 import math
 import time
 
+import torch.cuda
 from pero_ocr.utils import compose_path
 from .layout import PageLayout, RegionLayout, TextLine
 from pero_ocr.document_ocr import crop_engine as cropper
-from pero_ocr.ocr_engine import line_ocr_engine
+from pero_ocr.ocr_engine.pytorch_ocr_engine import PytorchEngineLineOCR
 from pero_ocr.layout_engines.simple_region_engine import SimpleThresholdRegion
 from pero_ocr.layout_engines.simple_baseline_engine import EngineLineDetectorSimple
 from pero_ocr.layout_engines.cnn_layout_engine import LayoutEngine, LineFilterEngine
@@ -19,18 +21,21 @@ from pero_ocr.layout_engines.baseline_refiner import refine_baseline
 from pero_ocr.layout_engines import layout_helpers as helpers
 
 
-def layout_parser_factory(config, config_path='', order=1):
+logger = logging.getLogger(__name__)
+
+
+def layout_parser_factory(config, device, config_path='', order=1):
     config = config['LAYOUT_PARSER_{}'.format(order)]
     if config['METHOD'] == 'REGION_WHOLE_PAGE':
         layout_parser = WholePageRegion(config, config_path=config_path)
     elif config['METHOD'] == 'REGION_SIMPLE_THRESHOLD':
         layout_parser = SimpleThresholdRegion(config, config_path=config_path)
     elif config['METHOD'] == 'LAYOUT_CNN':
-        layout_parser = LayoutExtractor(config, config_path=config_path)
+        layout_parser = LayoutExtractor(config, device, config_path=config_path)
     elif config['METHOD'] == 'LINES_SIMPLE_THRESHOLD':
         layout_parser = TextlineExtractorSimple(config, config_path=config_path)
     elif config['METHOD'] == 'LINE_FILTER':
-        layout_parser = LineFilter(config, config_path=config_path)
+        layout_parser = LineFilter(config, device, config_path=config_path)
     elif config['METHOD'] == 'LINE_POSTPROCESSING':
         layout_parser = LinePostprocessor(config, config_path=config_path)
     elif config['METHOD'] == 'LAYOUT_POSTPROCESSING':
@@ -49,61 +54,104 @@ def line_cropper_factory(config, config_path=''):
     return LineCropper(config, config_path=config_path)
 
 
-def ocr_factory(config, config_path=''):
+def ocr_factory(config, device, config_path=''):
     config = config['OCR']
-    return PageOCR(config, config_path=config_path)
+    return PageOCR(config, device, config_path=config_path)
 
 
-def page_decoder_factory(config, config_path=''):
+def page_decoder_factory(config, device, config_path=''):
     from pero_ocr.decoding import decoding_itf
     ocr_chars = decoding_itf.get_ocr_charset(compose_path(config['OCR']['OCR_JSON'], config_path))
-    decoder = decoding_itf.decoder_factory(config['DECODER'], ocr_chars, allow_no_decoder=False, use_gpu=True,
-                                           config_path=config_path)
+
+    use_cpu = config['DECODER'].getboolean('USE_CPU')
+    device = device if not use_cpu else torch.device("cpu")
+
+    decoder = decoding_itf.decoder_factory(config['DECODER'], ocr_chars, device, allow_no_decoder=False, config_path=config_path)
     confidence_threshold = config['DECODER'].getfloat('CONFIDENCE_THRESHOLD', fallback=math.inf)
-    return PageDecoder(decoder, line_confidence_threshold=confidence_threshold)
+    carry_h_over = config['DECODER'].getboolean('CARRY_H_OVER')
+    return PageDecoder(decoder, line_confidence_threshold=confidence_threshold, carry_h_over=carry_h_over)
 
 
 class MissingLogits(Exception):
     pass
 
 
+def line_confident_enough(logits, confidence_threshold):
+    log_probs = logits - np.logaddexp.reduce(logits, axis=1)[:, np.newaxis]
+    best_probs = np.max(log_probs, axis=-1)
+    worst_best_prob = np.exp(np.min(best_probs))
+
+    return worst_best_prob > confidence_threshold
+
+
+def prepare_dense_logits(line):
+    if line.logits is None:
+        raise MissingLogits(f"Line {line.id} has {line.logits} in place of logits")
+
+    return line.get_full_logprobs()
+
+
 class PageDecoder:
-    def __init__(self, decoder, line_confidence_threshold=None):
+    def __init__(self, decoder, line_confidence_threshold=None, carry_h_over=False):
         self.decoder = decoder
         self.line_confidence_threshold = line_confidence_threshold
         self.lines_examined = 0
         self.lines_decoded = 0
         self.seconds_decoding = 0.0
+        self.continue_lines = carry_h_over
+
+        self.last_h = None
+        self.last_line = None
 
     def process_page(self, page_layout: PageLayout):
+        self.last_h = None
         for line in page_layout.lines_iterator():
-            self.lines_examined += 1
-            logits = self.prepare_dense_logits(line)
-            if self.line_confidence_threshold is not None:
-                if self.line_confident_enough(logits):
-                    continue
-
-            t0 = time.time()
-            hypotheses = self.decoder(logits)
-            self.seconds_decoding += time.time() - t0
-            self.lines_decoded += 1
-            if hypotheses is not None:
-                line.transcription = hypotheses.best_hyp()
+            try:
+                line.transcription = self.decode_line(line)
+            except Exception:
+                logger.error(f'Failed to process line {line.id} of page {page_layout.id}. The page has been processed no further.', exc_info=True)
 
         return page_layout
 
-    def prepare_dense_logits(self, line):
-        if line.logits is None:
-            raise MissingLogits(f"Line {line.id} has {line.logits} in place of logits")
+    def decode_line(self, line):
+        self.lines_examined += 1
 
-        return line.get_full_logprobs()
+        logits = prepare_dense_logits(line)
+        if self.line_confidence_threshold is not None:
+            if line_confident_enough(logits, self.line_confidence_threshold):
+                self.last_h = None
+                self.last_line = line.transcription
+                return line.transcription
 
-    def line_confident_enough(self, logits):
-        log_probs = logits - np.logaddexp.reduce(logits, axis=1)[:, np.newaxis]
-        best_probs = np.max(log_probs, axis=-1)
-        worst_best_prob = np.exp(np.min(best_probs))
+        t0 = time.time()
+        if self.continue_lines:
+            if not self.last_h and self.last_line:
+                self.last_h = self.decoder._lm.initial_h_from_line(self.last_line)
 
-        return worst_best_prob > self.line_confidence_threshold
+            hypotheses, last_h = self.decoder(logits, return_h=True, init_h=self.last_h)
+            last_h = self.decoder._lm.add_line_end(last_h)
+            self.last_h = last_h
+        else:
+            hypotheses = self.decoder(logits)
+
+        self.seconds_decoding += time.time() - t0
+        self.lines_decoded += 1
+
+        transcription = hypotheses.best_hyp()
+        self.last_line = transcription
+
+        return transcription
+
+    def decoding_summary(self):
+        if self.lines_examined == 0:
+            return 'This PageDecoder has not processed a single line yet'
+
+        if self.lines_decoded == 0:
+            return f'Processed {self.lines_examined} lines, but none required actual decoding'
+
+        decoded_pct = 100.0 * self.lines_decoded / self.lines_examined
+        ms_per_line_decoded = 1000.0 * self.seconds_decoding / self.lines_decoded
+        return f'Ran on {self.lines_examined}, decoded {self.lines_decoded} lines ({decoded_pct:.1f} %) in {self.seconds_decoding:.2f}s ({ms_per_line_decoded:.1f}ms per line)'
 
 
 class WholePageRegion(object):
@@ -150,7 +198,7 @@ class TextlineExtractorSimple(object):
 
 
 class LayoutExtractor(object):
-    def __init__(self, config, config_path=''):
+    def __init__(self, config, device, config_path=''):
         self.detect_regions = config.getboolean('DETECT_REGIONS')
         self.detect_lines = config.getboolean('DETECT_LINES')
         self.detect_straight_lines_in_regions = config.getboolean('DETECT_STRAIGHT_LINES_IN_REGIONS')
@@ -158,18 +206,23 @@ class LayoutExtractor(object):
         self.adjust_heights = config.getboolean('ADJUST_HEIGHTS')
         self.multi_orientation = config.getboolean('MULTI_ORIENTATION')
         self.adjust_baselines = config.getboolean('ADJUST_BASELINES')
+
+        use_cpu = config.getboolean('USE_CPU')
+        self.device = device if not use_cpu else torch.device("cpu")
+
         self.engine = LayoutEngine(
-            framework=config.get('FRAMEWORK', fallback='tf'),
             model_path=compose_path(config['MODEL_PATH'], config_path),
+            device=self.device,
             downsample=config.getint('DOWNSAMPLE'),
             adaptive_downsample=config.getboolean('ADAPTIVE_DOWNSAMPLE', fallback=True),
-            pad=config.getint('PAD'),
-            use_cpu=config.getboolean('USE_CPU'),
             detection_threshold=config.getfloat('DETECTION_THRESHOLD'),
             max_mp=config.getfloat('MAX_MEGAPIXELS'),
-            gpu_fraction=config.getfloat('GPU_FRACTION')
+            line_end_weight=config.getfloat('LINE_END_WEIGHT', fallback=1.0),
+            vertical_line_connection_range=config.getint('VERTICAL_LINE_CONNECTION_RANGE', fallback=5),
+            smooth_line_predictions=config.getboolean('SMOOTH_LINE_PREDICTIONS', fallback=True),
+            paragraph_line_threshold=config.getfloat('PARAGRAPH_LINE_THRESHOLD', fallback=0.3),
         )
-        self.pool = Pool()
+        self.pool = Pool(1)
 
     def process_page(self, img, page_layout: PageLayout):
         if self.detect_regions or self.detect_lines:
@@ -245,18 +298,19 @@ class LayoutExtractor(object):
 
 
 class LineFilter(object):
-    def __init__(self, config, config_path):
+    def __init__(self, config, device, config_path):
         self.filter_directions = config.getboolean('FILTER_DIRECTIONS')
         self.filter_incomplete_pages = config.getboolean('FILTER_INCOMPLETE_PAGES')
         self.filter_pages_with_short_lines = config.getboolean('FILTER_PAGES_WITH_SHORT_LINES')
         self.length_threshold = config.getint('LENGTH_THRESHOLD')
 
+        use_cpu = config.getboolean('USE_CPU')
+        self.device = device if not use_cpu else torch.device("cpu")
+
         if self.filter_directions:
             self.engine = LineFilterEngine(
                 model_path=compose_path(config['MODEL_PATH'], config_path),
-                framework=config['FRAMEWORK'],
-                gpu_fraction=config.getfloat('GPU_FRACTION'),
-                use_cpu=config.getboolean('USE_CPU')
+                device=self.device
             )
 
     def process_page(self, img, page_layout: PageLayout):
@@ -336,15 +390,24 @@ class LineCropper(object):
                 print(f"WARNING: Failed to crop line {line.id} in page {page_layout.id}. Probably contain vertical line. Contanct Olda Kodym to fix this bug!")
         return page_layout
 
+    def crop_lines(self, img, lines: list):
+        for line in lines:
+            try:
+                line.crop = self.crop_engine.crop(
+                    img, line.baseline, line.heights)
+            except ValueError:
+                line.crop = np.zeros(
+                    (self.crop_engine.line_height, self.crop_engine.line_height, 3))
+                print(f"WARNING: Failed to crop line {line.id}. Probably contain vertical line. Contanct Olda Kodym to fix this bug!")
+
 
 class PageOCR(object):
-    def __init__(self, config, config_path=''):
+    def __init__(self, config, device, config_path=''):
         json_file = compose_path(config['OCR_JSON'], config_path)
-        if 'METHOD' in config and config['METHOD'] == 'pytorch_ocr':
-            from pero_ocr.ocr_engine.pytorch_ocr_engine import PytorchEngineLineOCR
-            self.ocr_engine = PytorchEngineLineOCR(json_file, gpu_id=0)
-        else:
-            self.ocr_engine = line_ocr_engine.EngineLineOCR(json_file, gpu_id=0)
+        use_cpu = config.getboolean('USE_CPU')
+
+        self.device = device if not use_cpu else torch.device("cpu")
+        self.ocr_engine = PytorchEngineLineOCR(json_file, self.device)
 
     def process_page(self, img, page_layout: PageLayout):
         for line in page_layout.lines_iterator():
@@ -378,7 +441,7 @@ def get_prob(best_ids, best_probs):
 
 
 class PageParser(object):
-    def __init__(self, config, config_path=''):
+    def __init__(self, config, device, config_path='', ):
         self.run_layout_parser = config['PAGE_PARSER'].getboolean('RUN_LAYOUT_PARSER', fallback=False)
         self.run_line_cropper = config['PAGE_PARSER'].getboolean('RUN_LINE_CROPPER', fallback=False)
         self.run_ocr = config['PAGE_PARSER'].getboolean('RUN_OCR', fallback=False)
@@ -391,35 +454,41 @@ class PageParser(object):
         self.ocr = None
         self.decoder = None
 
+        self.device = device
+
         if self.run_layout_parser:
             self.layout_parsers = []
             for i in range(1, 10):
                 if config.has_section('LAYOUT_PARSER_{}'.format(i)):
-                    self.layout_parsers.append(layout_parser_factory(config, config_path=config_path, order=i))
+                    self.layout_parsers.append(layout_parser_factory(config, device, config_path=config_path, order=i))
         if self.run_line_cropper:
             self.line_cropper = line_cropper_factory(config, config_path=config_path)
         if self.run_ocr:
-            self.ocr = ocr_factory(config, config_path=config_path)
+            self.ocr = ocr_factory(config, device, config_path=config_path)
         if self.run_decoder:
-            self.decoder = page_decoder_factory(config, config_path=config_path)
+            self.decoder = page_decoder_factory(config, device, config_path=config_path)
 
     @staticmethod
-    def compute_line_confidence(line, threshold):
+    def compute_line_confidence(line, threshold=None):
         logits = line.get_dense_logits()
         log_probs = logits - np.logaddexp.reduce(logits, axis=1)[:, np.newaxis]
         best_ids = np.argmax(log_probs, axis=-1)
         best_probs = np.exp(np.max(log_probs, axis=-1))
         worst_best_prob = get_prob(best_ids, best_probs)
-        print(worst_best_prob, np.sum(np.exp(best_probs) < threshold), best_probs.shape, np.nonzero(np.exp(best_probs) < threshold))
+        # print(worst_best_prob, np.sum(np.exp(best_probs) < threshold), best_probs.shape, np.nonzero(np.exp(best_probs) < threshold))
         # for i in np.nonzero(np.exp(best_probs) < threshold)[0]:
         #     print(best_probs[i-1:i+2], best_ids[i-1:i+2])
 
         return worst_best_prob
 
+    def update_confidences(self, page_layout):
+        for line in page_layout.lines_iterator():
+            if line.logits is not None:
+                line.transcription_confidence = self.compute_line_confidence(line)
+
     def filter_confident_lines(self, page_layout):
         for region in page_layout.regions:
-            region.lines = [line for line in region.lines
-                            if PageParser.compute_line_confidence(line, self.filter_confident_lines_threshold) > self.filter_confident_lines_threshold]
+            region.lines = [line for line in region.lines if line.transcription_confidence > self.filter_confident_lines_threshold]
         return page_layout
 
     def process_page(self, image, page_layout):
@@ -432,6 +501,9 @@ class PageParser(object):
             page_layout = self.ocr.process_page(image, page_layout)
         if self.run_decoder:
             page_layout = self.decoder.process_page(page_layout)
+
+        self.update_confidences(page_layout)
+
         if self.filter_confident_lines_threshold > 0:
             page_layout = self.filter_confident_lines(page_layout)
 
